@@ -27,7 +27,7 @@
  * list of people on the gedit Team.  
  * See the ChangeLog files for a list of changes. 
  *
- * $Id: gedit-document.c 6985 2009-04-11 12:59:08Z pborelli $
+ * $Id$
  */
 
 #ifdef HAVE_CONFIG_H
@@ -44,7 +44,6 @@
 #include "gedit-document.h"
 #include "gedit-debug.h"
 #include "gedit-utils.h"
-#include "gedit-metadata-manager.h"
 #include "gedit-language-manager.h"
 #include "gedit-style-scheme-manager.h"
 #include "gedit-document-loader.h"
@@ -53,6 +52,12 @@
 #include "gedit-enum-types.h"
 #include "gedittextregion.h"
 
+#ifdef G_OS_WIN32
+#include "gedit-metadata-manager.h"
+#else
+#define METADATA_QUERY "metadata::*"
+#endif
+
 #undef ENABLE_PROFILE 
 
 #ifdef ENABLE_PROFILE
@@ -82,8 +87,6 @@ static void	gedit_document_save_real	(GeditDocument          *doc,
 						 const gchar            *uri,
 						 const GeditEncoding    *encoding,
 						 GeditDocumentSaveFlags  flags);
-static void	gedit_document_set_readonly	(GeditDocument *doc,
-						 gboolean       readonly);
 static void	to_search_region_range 		(GeditDocument *doc,
 						 GtkTextIter   *start, 
 						 GtkTextIter   *end);
@@ -98,27 +101,25 @@ static void	delete_range_cb 		(GeditDocument *doc,
 			     
 struct _GeditDocumentPrivate
 {
-	gint	     readonly : 1;
-	gint	     last_save_was_manually : 1; 	
-	gint	     language_set_by_user : 1;
-	gint         stop_cursor_moved_emission : 1;
-	gint         dispose_has_run : 1;
-
 	gchar	    *uri;
 	gint 	     untitled_number;
+	gchar       *short_name;
+
+	GFileInfo   *metadata_info;
 
 	const GeditEncoding *encoding;
 
 	gchar	    *content_type;
 
-	time_t       mtime;
-
+	GTimeVal     mtime;
 	GTimeVal     time_of_last_save_or_load;
 
 	guint        search_flags;
 	gchar       *search_text;
 	gint	     num_of_lines_search_text;
 
+	GeditDocumentNewlineType newline_type;
+
 	/* Temp data while loading */
 	GeditDocumentLoader *loader;
 	gboolean             create; /* Create file if uri points 
@@ -129,13 +130,19 @@ struct _GeditDocumentPrivate
 	/* Saving stuff */
 	GeditDocumentSaver *saver;
 
-	/* Search highlighting support variables */	
+	/* Search highlighting support variables */
 	GeditTextRegion *to_search_region;
 	GtkTextTag      *found_tag;
-	
+
 	/* Mount operation factory */
 	GeditMountOperationFactory  mount_operation_factory;
 	gpointer		    mount_operation_userdata;
+
+	gint readonly : 1;
+	gint last_save_was_manually : 1; 
+	gint language_set_by_user : 1;
+	gint stop_cursor_moved_emission : 1;
+	gint dispose_has_run : 1;
 };
 
 enum {
@@ -148,7 +155,8 @@ enum {
 	PROP_READ_ONLY,
 	PROP_ENCODING,
 	PROP_CAN_SEARCH_AGAIN,
-	PROP_ENABLE_SEARCH_HIGHLIGHTING
+	PROP_ENABLE_SEARCH_HIGHLIGHTING,
+	PROP_NEWLINE_TYPE
 };
 
 enum {
@@ -228,19 +236,7 @@ gedit_document_dispose (GObject *object)
 	{
 		GtkTextIter iter;
 		gchar *position;
-
-		gtk_text_buffer_get_iter_at_mark (
-				GTK_TEXT_BUFFER (doc),			
-				&iter,
-				gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (doc)));
-
-		position = g_strdup_printf ("%d", 
-					    gtk_text_iter_get_offset (&iter));
-
-		gedit_metadata_manager_set (doc->priv->uri,
-					    "position",
-					    position);
-		g_free (position);
+		const gchar *language = NULL;
 
 		if (doc->priv->language_set_by_user)
 		{
@@ -248,10 +244,28 @@ gedit_document_dispose (GObject *object)
 
 			lang = gedit_document_get_language (doc);
 
-			gedit_metadata_manager_set (doc->priv->uri,
-				    "language",
-				    (lang == NULL) ? "_NORMAL_" : gtk_source_language_get_id (lang));
+			if (lang == NULL)
+				language = "_NORMAL_";
+			else
+				language = gtk_source_language_get_id (lang);
 		}
+
+		gtk_text_buffer_get_iter_at_mark (
+				GTK_TEXT_BUFFER (doc),
+				&iter,
+				gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (doc)));
+
+		position = g_strdup_printf ("%d", 
+					    gtk_text_iter_get_offset (&iter));
+
+		if (language == NULL)
+			gedit_document_set_metadata (doc, GEDIT_METADATA_ATTRIBUTE_POSITION,
+						     position, NULL);
+		else
+			gedit_document_set_metadata (doc, GEDIT_METADATA_ATTRIBUTE_POSITION,
+						     position, GEDIT_METADATA_ATTRIBUTE_LANGUAGE,
+						     language, NULL);
+		g_free (position);
 	}
 
 	if (doc->priv->loader)
@@ -260,6 +274,12 @@ gedit_document_dispose (GObject *object)
 		doc->priv->loader = NULL;
 	}
 
+	if (doc->priv->metadata_info != NULL)
+	{
+		g_object_unref (doc->priv->metadata_info);
+		doc->priv->metadata_info = NULL;
+	}
+
 	doc->priv->dispose_has_run = TRUE;
 
 	G_OBJECT_CLASS (gedit_document_parent_class)->dispose (object);
@@ -325,6 +345,9 @@ gedit_document_get_property (GObject    *object,
 		case PROP_ENABLE_SEARCH_HIGHLIGHTING:
 			g_value_set_boolean (value, gedit_document_get_enable_search_highlighting (doc));
 			break;
+		case PROP_NEWLINE_TYPE:
+			g_value_set_enum (value, doc->priv->newline_type);
+			break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -344,6 +367,19 @@ gedit_document_set_property (GObject      *object,
 		case PROP_ENABLE_SEARCH_HIGHLIGHTING:
 			gedit_document_set_enable_search_highlighting (doc,
 								       g_value_get_boolean (value));
+			break;
+		case PROP_NEWLINE_TYPE:
+			gedit_document_set_newline_type (doc,
+							 g_value_get_enum (value));
+			break;
+		case PROP_SHORTNAME:
+			gedit_document_set_short_name_for_display (doc,
+			                                           g_value_get_string (value));
+			break;
+		case PROP_CONTENT_TYPE:
+			gedit_document_set_content_type (doc,
+			                                 g_value_get_string (value));
+			break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -396,7 +432,7 @@ gedit_document_class_init (GeditDocumentClass *klass)
 	object_class->dispose = gedit_document_dispose;
 	object_class->finalize = gedit_document_finalize;
 	object_class->get_property = gedit_document_get_property;
-	object_class->set_property = gedit_document_set_property;	
+	object_class->set_property = gedit_document_set_property;
 
 	buf_class->mark_set = gedit_document_mark_set;
 	buf_class->changed = gedit_document_changed;
@@ -417,7 +453,7 @@ gedit_document_class_init (GeditDocumentClass *klass)
 							      "Short Name",
 							      "The document's short name",
 							      NULL,
-							      G_PARAM_READABLE |
+							      G_PARAM_READWRITE |
 							      G_PARAM_STATIC_STRINGS));
 
 	g_object_class_install_property (object_class, PROP_CONTENT_TYPE,
@@ -425,7 +461,7 @@ gedit_document_class_init (GeditDocumentClass *klass)
 							      "Content Type",
 							      "The document's Content Type",
 							      NULL,
-							      G_PARAM_READABLE |
+							      G_PARAM_READWRITE |
 							      G_PARAM_STATIC_STRINGS));
 
 	g_object_class_install_property (object_class, PROP_MIME_TYPE,
@@ -468,6 +504,22 @@ gedit_document_class_init (GeditDocumentClass *klass)
 							       G_PARAM_READWRITE |
 							       G_PARAM_STATIC_STRINGS));
 
+	/**
+	 * GeditDocument:newline-type:
+	 *
+	 * The :newline-type property determines what is considered
+	 * as a line ending when saving the document
+	 */
+	g_object_class_install_property (object_class, PROP_NEWLINE_TYPE,
+	                                 g_param_spec_enum ("newline-type",
+	                                                    "Newline type",
+	                                                    "The accepted types of line ending",
+	                                                    GEDIT_TYPE_DOCUMENT_NEWLINE_TYPE,
+	                                                    GEDIT_DOCUMENT_NEWLINE_TYPE_LF,
+	                                                    G_PARAM_READWRITE |
+	                                                    G_PARAM_STATIC_NAME |
+	                                                    G_PARAM_STATIC_BLURB));
+
 	/* This signal is used to update the cursor position is the statusbar,
 	 * it's emitted either when the insert mark is moved explicitely or
 	 * when the buffer changes (insert/delete).
@@ -625,9 +677,9 @@ set_language (GeditDocument     *doc,
 
 	if (set_by_user && (doc->priv->uri != NULL))
 	{
-		gedit_metadata_manager_set (doc->priv->uri,
-			    "language",
-			    (lang == NULL) ? "_NORMAL_" : gtk_source_language_get_id (lang));
+		gedit_document_set_metadata (doc, GEDIT_METADATA_ATTRIBUTE_LANGUAGE,
+			(lang == NULL) ? "_NORMAL_" : gtk_source_language_get_id (lang),
+			NULL);
 	}
 
 	doc->priv->language_set_by_user = set_by_user;
@@ -653,9 +705,8 @@ set_encoding (GeditDocument       *doc,
 
 		charset = gedit_encoding_get_charset (encoding);
 
-		gedit_metadata_manager_set (doc->priv->uri,
-					    "encoding",
-					    charset);
+		gedit_document_set_metadata (doc, GEDIT_METADATA_ATTRIBUTE_ENCODING,
+					     charset, NULL);
 	}
 
 	g_object_notify (G_OBJECT (doc), "encoding");
@@ -689,6 +740,79 @@ get_default_style_scheme (void)
 	return def_style;
 }
 
+static GtkSourceLanguage *
+guess_language (GeditDocument *doc,
+		const gchar   *content_type)
+{
+	gchar *data;
+	GtkSourceLanguage *language = NULL;
+
+	data = gedit_document_get_metadata (doc, GEDIT_METADATA_ATTRIBUTE_LANGUAGE);
+
+	if (data != NULL)
+	{
+		gedit_debug_message (DEBUG_DOCUMENT, "Language from metadata: %s", data);
+
+		if (strcmp (data, "_NORMAL_") != 0)
+		{
+			language = gtk_source_language_manager_get_language (
+						gedit_get_language_manager (),
+						data);
+		}
+
+		g_free (data);
+	}
+	else
+	{
+		GFile *file;
+		gchar *basename = NULL;
+
+		file = gedit_document_get_location (doc);
+		gedit_debug_message (DEBUG_DOCUMENT, "Sniffing Language");
+
+		if (file)
+		{
+			basename = g_file_get_basename (file);
+		}
+		else if (doc->priv->short_name != NULL)
+		{
+			basename = g_strdup (doc->priv->short_name);
+		}
+
+		language = gtk_source_language_manager_guess_language (
+					gedit_get_language_manager (),
+					basename,
+					content_type);
+
+		g_free (basename);
+
+		if (file != NULL)
+		{
+			g_object_unref (file);
+		}
+	}
+
+	return language;
+}
+
+static void
+on_content_type_changed (GeditDocument *doc,
+			 GParamSpec    *pspec,
+			 gpointer       useless)
+{
+	if (!doc->priv->language_set_by_user)
+	{
+		GtkSourceLanguage *language;
+
+		language = guess_language (doc, doc->priv->content_type);
+
+		gedit_debug_message (DEBUG_DOCUMENT, "Language: %s",
+				     language != NULL ? gtk_source_language_get_name (language) : "None");
+
+		set_language (doc, language, FALSE);
+	}
+}
+
 static void
 gedit_document_init (GeditDocument *doc)
 {
@@ -701,6 +825,8 @@ gedit_document_init (GeditDocument *doc)
 	doc->priv->uri = NULL;
 	doc->priv->untitled_number = get_untitled_number ();
 
+	doc->priv->metadata_info = NULL;
+
 	doc->priv->content_type = g_content_type_from_mime_type ("text/plain");
 
 	doc->priv->readonly = FALSE;
@@ -712,12 +838,15 @@ gedit_document_init (GeditDocument *doc)
 
 	doc->priv->dispose_has_run = FALSE;
 
-	doc->priv->mtime = 0;
+	doc->priv->mtime.tv_sec = 0;
+	doc->priv->mtime.tv_usec = 0;
 
 	g_get_current_time (&doc->priv->time_of_last_save_or_load);
 
 	doc->priv->encoding = gedit_encoding_get_utf8 ();
 
+	doc->priv->newline_type = GEDIT_DOCUMENT_NEWLINE_TYPE_DEFAULT;
+
 	gtk_source_buffer_set_max_undo_levels (GTK_SOURCE_BUFFER (doc), 
 					       gedit_prefs_manager_get_undo_actions_limit ());
 
@@ -740,7 +869,12 @@ gedit_document_init (GeditDocument *doc)
 	g_signal_connect_after (doc, 
 			  	"delete-range",
 			  	G_CALLBACK (delete_range_cb),
-			  	NULL);		
+			  	NULL);
+
+	g_signal_connect (doc,
+			  "notify::content-type",
+			  G_CALLBACK (on_content_type_changed),
+			  NULL);
 }
 
 GeditDocument *
@@ -751,57 +885,86 @@ gedit_document_new (void)
 	return GEDIT_DOCUMENT (g_object_new (GEDIT_TYPE_DOCUMENT, NULL));
 }
 
-static GtkSourceLanguage *
-guess_language (const gchar *uri,
-		const gchar *content_type)
-
+static void
+set_content_type (GeditDocument *doc,
+		  const gchar   *content_type)
 {
-	gchar *data;
-	GtkSourceLanguage *language = NULL;
+	gedit_debug (DEBUG_DOCUMENT);
 
-	data = gedit_metadata_manager_get (uri, "language");
+	g_free (doc->priv->content_type);
 
-	if (data != NULL)
+	if (content_type != NULL)
 	{
-		gedit_debug_message (DEBUG_DOCUMENT, "Language from metadata: %s", data);
-
-		if (strcmp (data, "_NORMAL_") != 0)
-		{
-			language = gtk_source_language_manager_get_language (
-						gedit_get_language_manager (),
-						data);
-		}
-
-		g_free (data);
+		doc->priv->content_type = g_strdup (content_type);
 	}
 	else
 	{
 		GFile *file;
-		gchar *basename;
+		gchar *guessed_type = NULL;
 
-		gedit_debug_message (DEBUG_DOCUMENT, "Sniffing Language");
+		/* If content type is null, we guess from the filename */
+		file = gedit_document_get_location (doc);
+		if (file != NULL)
+		{
+			gchar *basename;
 
-		file = g_file_new_for_uri (uri);
-		basename = g_file_get_basename (file);
+			basename = g_file_get_basename (file);
+			guessed_type = g_content_type_guess (basename, NULL, 0, NULL);
 
-		language = gtk_source_language_manager_guess_language (
-					gedit_get_language_manager (),
-					basename,
-					content_type);
+			g_free (basename);
+			g_object_unref (file);
+		}
 
-		g_free (basename);
-		g_object_unref (file);
+		if (guessed_type == NULL || g_content_type_is_unknown (guessed_type))
+		{
+			g_free (guessed_type);
+			guessed_type = g_content_type_from_mime_type ("text/plain");
+		}
+
+		doc->priv->content_type = guessed_type;
 	}
 
-	return language;
+	g_object_notify (G_OBJECT (doc), "content-type");
 }
 
-/* If content type is null, we guess from the filename */
-/* If uri is null, we only set the content-type */
+void
+gedit_document_set_content_type (GeditDocument *doc,
+                                 const gchar   *content_type)
+{
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+
+	set_content_type (doc, content_type);
+}
+
+#ifndef G_OS_WIN32
+static void
+query_info_cb (GFile         *source,
+	       GAsyncResult  *res,
+	       GeditDocument *doc)
+{
+	GError *error = NULL;
+
+	doc->priv->metadata_info = g_file_query_info_finish (source,
+							     res,
+							     &error);
+
+	if (error != NULL)
+	{
+		if (error->code != G_FILE_ERROR_ISDIR)
+			g_warning ("%s", error->message);
+		g_error_free (error);
+		
+		return;
+	}
+
+	
+	on_content_type_changed (doc, NULL, NULL);
+}
+#endif
+
 static void
 set_uri (GeditDocument *doc,
-	 const gchar   *uri,
-	 const gchar   *content_type)
+	 const gchar   *uri)
 {
 	gedit_debug (DEBUG_DOCUMENT);
 
@@ -822,38 +985,31 @@ set_uri (GeditDocument *doc,
 		}
 	}
 
-	g_free (doc->priv->content_type);
+#ifndef G_OS_WIN32
+	GFile *location;
 
-	if (content_type != NULL)
+	/* Get the GFileInfo async so we can set the language from the metadata */
+	location = gedit_document_get_location (doc);
+	
+	if (location != NULL)
 	{
-		doc->priv->content_type = g_strdup (content_type);
-	}
-	else
-	{
-		if (doc->priv->uri != NULL)
-		{
-			doc->priv->content_type = g_content_type_guess (doc->priv->uri, NULL, 0, NULL);
-		}
-		else
-		{
-			doc->priv->content_type = g_content_type_from_mime_type ("text/plain");
-		}
-	}
-
-	if (!doc->priv->language_set_by_user)
-	{
-		GtkSourceLanguage *language;
-
-		language = guess_language (doc->priv->uri, doc->priv->content_type);
-
-		gedit_debug_message (DEBUG_DOCUMENT, "Language: %s",
-				     language != NULL ? gtk_source_language_get_name (language) : "None");
-
-		set_language (doc, language, FALSE);
+		g_file_query_info_async (location,
+					 METADATA_QUERY,
+					 G_FILE_QUERY_INFO_NONE,
+					 G_PRIORITY_DEFAULT,
+					 NULL,
+					 (GAsyncReadyCallback) query_info_cb,
+					 doc);
+		g_object_unref (location);
 	}
+#endif
 
 	g_object_notify (G_OBJECT (doc), "uri");
-	g_object_notify (G_OBJECT (doc), "shortname");
+
+	if (doc->priv->short_name == NULL)
+	{
+		g_object_notify (G_OBJECT (doc), "shortname");
+	}
 }
 
 GFile *
@@ -876,14 +1032,15 @@ void
 gedit_document_set_uri (GeditDocument *doc,
 			const gchar   *uri)
 {
-	set_uri (doc, uri, doc->priv->content_type);
+	set_uri (doc, uri);
+	set_content_type (doc, NULL);
 }
 
 /* Never returns NULL */
 gchar *
 gedit_document_get_uri_for_display (GeditDocument *doc)
 {
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), "");
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), g_strdup (""));
 
 	if (doc->priv->uri == NULL)
 		return g_strdup_printf (_("Unsaved Document %d"),
@@ -896,15 +1053,29 @@ gedit_document_get_uri_for_display (GeditDocument *doc)
 gchar *
 gedit_document_get_short_name_for_display (GeditDocument *doc)
 {
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), "");
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), g_strdup (""));
 
-	if (doc->priv->uri == NULL)
+	if (doc->priv->short_name != NULL)
+		return g_strdup (doc->priv->short_name);
+	else if (doc->priv->uri == NULL)
 		return g_strdup_printf (_("Unsaved Document %d"),
 					doc->priv->untitled_number);
 	else
 		return gedit_utils_basename_for_display (doc->priv->uri);
 }
 
+void
+gedit_document_set_short_name_for_display (GeditDocument *doc,
+                                           const gchar   *short_name)
+{
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+
+	g_free (doc->priv->short_name);
+	doc->priv->short_name = g_strdup (short_name);
+
+	g_object_notify (G_OBJECT (doc), "shortname");
+}
+
 gchar *
 gedit_document_get_content_type (GeditDocument *doc)
 {
@@ -919,7 +1090,7 @@ gedit_document_get_mime_type (GeditDocument *doc)
 {
 	gchar *mime_type = NULL;
 
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), "text/plain");
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), g_strdup ("text/plain"));
 
 	if ((doc->priv->content_type != NULL) &&
 	    (!g_content_type_is_unknown (doc->priv->content_type)))
@@ -945,9 +1116,16 @@ set_readonly (GeditDocument *doc,
 	doc->priv->readonly = readonly;
 }
 
-static void
-gedit_document_set_readonly (GeditDocument *doc,
-			     gboolean       readonly)
+/**
+ * gedit_document_set_readonly:
+ * @doc: a #GeditDocument
+ * @readonly: %TRUE to se the document as read-only
+ *
+ * If @readonly is %TRUE sets @doc as read-only.
+ */
+void
+_gedit_document_set_readonly (GeditDocument *doc,
+			      gboolean       readonly)
 {
 	gedit_debug (DEBUG_DOCUMENT);
 
@@ -996,7 +1174,9 @@ _gedit_document_check_externally_modified (GeditDocument *doc)
 	g_file_info_get_modification_time (info, &timeval);
 	g_object_unref (info);
 	
-	return timeval.tv_sec > doc->priv->mtime;
+	return (timeval.tv_sec > doc->priv->mtime.tv_sec) ||
+	       (timeval.tv_sec == doc->priv->mtime.tv_sec && 
+	       timeval.tv_usec > doc->priv->mtime.tv_usec);
 }
 
 static void
@@ -1016,26 +1196,45 @@ document_loader_loaded (GeditDocumentLoader *loader,
 			GeditDocument       *doc)
 {
 	/* load was successful */
-	if (error == NULL)
+	if (error == NULL || error->code == GEDIT_DOCUMENT_ERROR_CONVERSION_FALLBACK)
 	{
 		GtkTextIter iter;
-		const gchar *content_type;
+		GFileInfo *info;
+		const gchar *content_type = NULL;
+		gboolean read_only = FALSE;
+		GTimeVal mtime = {0, 0};
 
-		content_type = gedit_document_loader_get_content_type (loader);
+		info = gedit_document_loader_get_info (loader);
 
-		doc->priv->mtime = gedit_document_loader_get_mtime (loader);
+		if (info)
+		{
+			if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE))
+				content_type = g_file_info_get_attribute_string (info,
+										 G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE);
+
+			if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_TIME_MODIFIED))
+				g_file_info_get_modification_time (info, &mtime);
+
+			if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE))
+				read_only = !g_file_info_get_attribute_boolean (info,
+										G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE);
+		}
+
+		doc->priv->mtime = mtime;
+
+		set_readonly (doc, read_only);
 
 		g_get_current_time (&doc->priv->time_of_last_save_or_load);
 
-		set_readonly (doc,
-		      gedit_document_loader_get_readonly (loader));
-
 		set_encoding (doc, 
 			      gedit_document_loader_get_encoding (loader),
 			      (doc->priv->requested_encoding != NULL));
-		      
-		/* We already set the uri */
-		set_uri (doc, NULL, content_type);
+
+
+		set_content_type (doc, content_type);
+
+		gedit_document_set_newline_type (doc,
+		                                 gedit_document_loader_get_newline_type (loader));
 
 		/* move the cursor at the requested line if any */
 		if (doc->priv->requested_line_pos > 0)
@@ -1051,8 +1250,7 @@ document_loader_loaded (GeditDocumentLoader *loader,
 			gchar *pos;
 			gint offset;
 
-			pos = gedit_metadata_manager_get (doc->priv->uri,
-							  "position");
+			pos = gedit_document_get_metadata (doc, GEDIT_METADATA_ATTRIBUTE_POSITION);
 
 			offset = pos ? atoi (pos) : 0;
 			g_free (pos);
@@ -1115,10 +1313,16 @@ document_loader_loading (GeditDocumentLoader *loader,
 	}
 	else
 	{
-		goffset size;
+		goffset size = 0;
 		goffset read;
+		GFileInfo *info;
+
+		info = gedit_document_loader_get_info (loader);
+
+		if (info && g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_STANDARD_SIZE))
+			size = g_file_info_get_attribute_uint64 (info,
+								 G_FILE_ATTRIBUTE_STANDARD_SIZE);
 
-		size = gedit_document_loader_get_file_size (loader);
 		read = gedit_document_loader_get_bytes_read (loader);
 
 		g_signal_emit (doc, 
@@ -1152,7 +1356,8 @@ gedit_document_load_real (GeditDocument       *doc,
 	doc->priv->requested_encoding = encoding;
 	doc->priv->requested_line_pos = line_pos;
 
-	set_uri (doc, uri, NULL);
+	set_uri (doc, uri);
+	set_content_type (doc, NULL);
 
 	gedit_document_loader_load (doc->priv->loader);
 }
@@ -1212,23 +1417,35 @@ document_saver_saving (GeditDocumentSaver *saver,
 		if (error == NULL)
 		{
 			const gchar *uri;
-			const gchar *content_type;
+			const gchar *content_type = NULL;
+			GTimeVal mtime = {0, 0};
+			GFileInfo *info;
 
 			uri = gedit_document_saver_get_uri (saver);
+			set_uri (doc, uri);
 
-			content_type = gedit_document_saver_get_content_type (saver);
+			info = gedit_document_saver_get_info (saver);
 
-			doc->priv->mtime = gedit_document_saver_get_mtime (saver);
+			if (info != NULL)
+			{
+				if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE))
+					content_type = g_file_info_get_attribute_string (info,
+											 G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE);
+
+				if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_TIME_MODIFIED))
+					g_file_info_get_modification_time (info, &mtime);
+			}
+
+			set_content_type (doc, content_type);
+			doc->priv->mtime = mtime;
 
 			g_get_current_time (&doc->priv->time_of_last_save_or_load);
 
-			gedit_document_set_readonly (doc, FALSE);
+			_gedit_document_set_readonly (doc, FALSE);
 
 			gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (doc),
 						      FALSE);
 
-			set_uri (doc, uri, content_type);
-
 			set_encoding (doc, 
 				      doc->priv->requested_encoding, 
 				      TRUE);
@@ -1270,7 +1487,9 @@ gedit_document_save_real (GeditDocument          *doc,
 	g_return_if_fail (doc->priv->saver == NULL);
 
 	/* create a saver, it will be destroyed once saving is complete */
-	doc->priv->saver = gedit_document_saver_new (doc, uri, encoding, flags);
+	doc->priv->saver = gedit_document_saver_new (doc, uri, encoding,
+						     doc->priv->newline_type,
+						     flags);
 
 	g_signal_connect (doc->priv->saver,
 			  "saving",
@@ -1280,7 +1499,7 @@ gedit_document_save_real (GeditDocument          *doc,
 	doc->priv->requested_encoding = encoding;
 
 	gedit_document_saver_save (doc->priv->saver,
-				   doc->priv->mtime);
+				   &doc->priv->mtime);
 }
 
 /**
@@ -1427,6 +1646,38 @@ gedit_document_goto_line (GeditDocument *doc,
 	return ret;
 }
 
+gboolean
+gedit_document_goto_line_offset (GeditDocument *doc,
+				 gint           line,
+				 gint           line_offset)
+{
+	gboolean ret = TRUE;
+	guint offset_count;
+	GtkTextIter iter;
+	
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), FALSE);
+	g_return_val_if_fail (line >= -1, FALSE);
+	g_return_val_if_fail (line_offset >= -1, FALSE);
+	
+	gtk_text_buffer_get_iter_at_line (GTK_TEXT_BUFFER (doc),
+					  &iter,
+					  line);
+
+	offset_count = gtk_text_iter_get_chars_in_line (&iter);
+	if (line_offset > offset_count)
+	{
+		ret = FALSE;
+	}
+	else
+	{
+		gtk_text_iter_set_line_offset (&iter, line_offset);
+	}
+	
+	gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc), &iter);
+
+	return ret;
+}
+
 static gint
 compute_num_of_lines (const gchar *text)
 {
@@ -1914,15 +2165,12 @@ sync_found_tag (GeditDocument *doc,
 				 &fg_set, &fg,
 				 &bg_set, &bg);
 
-	if (fg_set)
-		g_object_set (doc->priv->found_tag,
-			      "foreground-gdk", &fg,
-			      NULL);
-
-	if (bg_set)
-		g_object_set (doc->priv->found_tag,
-			      "background-gdk", &bg,
-			      NULL);
+	g_object_set (doc->priv->found_tag,
+		      "foreground-gdk", fg_set ? &fg : NULL,
+		      NULL);
+	g_object_set (doc->priv->found_tag,
+		      "background-gdk", bg_set ? &bg : NULL,
+		      NULL);
 }
 
 static void
@@ -2234,6 +2482,28 @@ gedit_document_get_enable_search_highlighting (GeditDocument *doc)
 	return (doc->priv->to_search_region != NULL);
 }
 
+void
+gedit_document_set_newline_type (GeditDocument           *doc,
+				 GeditDocumentNewlineType newline_type)
+{
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+
+	if (doc->priv->newline_type != newline_type)
+	{
+		doc->priv->newline_type = newline_type;
+
+		g_object_notify (G_OBJECT (doc), "newline-type");
+	}
+}
+
+GeditDocumentNewlineType
+gedit_document_get_newline_type (GeditDocument *doc)
+{
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), 0);
+
+	return doc->priv->newline_type;
+}
+
 void
 _gedit_document_set_mount_operation_factory (GeditDocument 	       *doc,
 					    GeditMountOperationFactory	callback,
@@ -2256,3 +2526,155 @@ _gedit_document_create_mount_operation (GeditDocument *doc)
 		return doc->priv->mount_operation_factory (doc, 
 						           doc->priv->mount_operation_userdata);
 }
+
+#ifdef G_OS_WIN32
+gchar *
+gedit_document_get_metadata (GeditDocument *doc,
+			     const gchar   *key)
+{
+	gchar *value = NULL;
+
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), NULL);
+	g_return_val_if_fail (key != NULL, NULL);
+
+	if (doc->priv->uri != NULL)
+		value = gedit_metadata_manager_get (doc->priv->uri, key);
+
+	return value;
+}
+
+void
+gedit_document_set_metadata (GeditDocument *doc,
+			     const gchar   *first_key,
+			     ...)
+{
+	const gchar *key;
+	const gchar *value;
+	va_list var_args;
+
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	g_return_if_fail (first_key != NULL);
+
+	va_start (var_args, first_key);
+
+	for (key = first_key; key; key = va_arg (var_args, const gchar *))
+	{
+		value = va_arg (var_args, const gchar *);
+		
+		gedit_metadata_manager_set (doc->priv->uri,
+					    key,
+					    value);
+	}
+
+	va_end (var_args);
+}
+
+#else
+
+/**
+ * gedit_document_get_metadata:
+ * @doc: a #GeditDocument
+ * @key: name of the key
+ *
+ * Gets the metadata assigned to @key.
+ *
+ * Returns: the value assigned to @key.
+ */
+gchar *
+gedit_document_get_metadata (GeditDocument *doc,
+			     const gchar   *key)
+{
+	gchar *value = NULL;
+
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT (doc), NULL);
+	g_return_val_if_fail (key != NULL, NULL);
+
+	if (doc->priv->metadata_info && g_file_info_has_attribute (doc->priv->metadata_info,
+								   key))
+	{
+		value = g_strdup (g_file_info_get_attribute_string (doc->priv->metadata_info,
+								    key));
+	}
+
+	return value;
+}
+
+static void
+set_attributes_cb (GObject      *source,
+		   GAsyncResult *res,
+		   gpointer      useless)
+{
+	g_file_set_attributes_finish (G_FILE (source),
+				      res,
+				      NULL,
+				      NULL);
+}
+
+/**
+ * gedit_document_set_metadata:
+ * @doc: a #GeditDocument
+ * @first_key: name of the first key to set
+ * @...: value for the first key, followed optionally by more key/value pairs,
+ * followed by %NULL.
+ *
+ * Sets metadata on a document.
+ */
+void
+gedit_document_set_metadata (GeditDocument *doc,
+			     const gchar   *first_key,
+			     ...)
+{
+	const gchar *key;
+	const gchar *value;
+	va_list var_args;
+	GFileInfo *info;
+	GFile *location;
+
+	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	g_return_if_fail (first_key != NULL);
+
+	info = g_file_info_new ();
+
+	va_start (var_args, first_key);
+
+	for (key = first_key; key; key = va_arg (var_args, const gchar *))
+	{
+		value = va_arg (var_args, const gchar *);
+		
+		if (value != NULL)
+		{
+			g_file_info_set_attribute_string (info,
+							  key, value);
+		}
+		else
+		{
+			/* Unset the key */
+			g_file_info_set_attribute (info, key,
+						   G_FILE_ATTRIBUTE_TYPE_INVALID,
+						   NULL);
+		}
+	}
+
+	va_end (var_args);
+
+	if (doc->priv->metadata_info != NULL)
+		g_file_info_copy_into (info, doc->priv->metadata_info);
+
+	location = gedit_document_get_location (doc);
+
+	if (location != NULL)
+	{
+		g_file_set_attributes_async (location,
+					     info,
+					     G_FILE_QUERY_INFO_NONE,
+					     G_PRIORITY_DEFAULT,
+					     NULL,
+					     set_attributes_cb,
+					     NULL);
+
+		g_object_unref (location);
+	}
+	
+	g_object_unref (info);
+}
+#endif
