@@ -39,8 +39,8 @@
 #include <gio/gio.h>
 
 #include "gedit-gio-document-loader.h"
+#include "gedit-document-output-stream.h"
 #include "gedit-smart-charset-converter.h"
-#include "gedit-convert.h"
 #include "gedit-prefs-manager.h"
 #include "gedit-debug.h"
 #include "gedit-utils.h"
@@ -49,6 +49,8 @@ typedef struct
 {
 	GeditGioDocumentLoader *loader;
 	GCancellable 	       *cancellable;
+
+	gssize			read;
 	gboolean		tried_mount;
 } AsyncData;
 
@@ -81,13 +83,12 @@ struct _GeditGioDocumentLoaderPrivate
 	/* Handle for remote files */
 	GCancellable 	 *cancellable;
 	GInputStream	 *stream;
+	GOutputStream    *output;
 	GeditSmartCharsetConverter *converter;
 
 	gchar             buffer[READ_CHUNK_SIZE];
 
 	GError           *error;
-
-	guint		  started_insert : 1;
 };
 
 G_DEFINE_TYPE(GeditGioDocumentLoader, gedit_gio_document_loader, GEDIT_TYPE_DOCUMENT_LOADER)
@@ -112,6 +113,12 @@ gedit_gio_document_loader_dispose (GObject *object)
 		priv->stream = NULL;
 	}
 
+	if (priv->output != NULL)
+	{
+		g_object_unref (priv->output);
+		priv->output = NULL;
+	}
+
 	if (priv->converter != NULL)
 	{
 		g_object_unref (priv->converter);
@@ -162,7 +169,6 @@ gedit_gio_document_loader_init (GeditGioDocumentLoader *gvloader)
 
 	gvloader->priv->converter = NULL;
 	gvloader->priv->error = NULL;
-	gvloader->priv->started_insert = FALSE;
 }
 
 static AsyncData *
@@ -230,25 +236,11 @@ get_metadata_encoding (GeditDocumentLoader *loader)
 static void
 remote_load_completed_or_failed (GeditGioDocumentLoader *gvloader, AsyncData *async)
 {
-	GeditDocumentLoader *loader;
-
-	loader = GEDIT_DOCUMENT_LOADER (gvloader);
-
-	if (async)
-		async_data_free (async);
-
-	if (gvloader->priv->stream)
-		g_input_stream_close_async (G_INPUT_STREAM (gvloader->priv->stream),
-					    G_PRIORITY_HIGH, NULL, NULL, NULL);
-
-	gedit_document_loader_loading (GEDIT_DOCUMENT_LOADER (gvloader),
+	gedit_document_loader_loading (GEDIT_DOCUMENT_LOADER (async->loader),
 				       TRUE,
-				       gvloader->priv->error);
+				       async->loader->priv->error);
 }
 
-/* prototype, because they call each other... isn't C lovely */
-static void	read_file_chunk		(AsyncData *async);
-
 static void
 async_failed (AsyncData *async, GError *error)
 {
@@ -257,110 +249,91 @@ async_failed (AsyncData *async, GError *error)
 }
 
 static void
-append_text_to_document (GeditDocumentLoader *loader,
-			 const gchar         *text,
-			 gint                 len)
+close_input_stream_ready_cb (GInputStream *stream,
+			     GAsyncResult  *res,
+			     AsyncData     *async)
 {
-	GeditDocument *doc = loader->document;
-	GtkTextIter end;
-
-	/* Insert text in the buffer */
-	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (doc), &end);
+	GError *error = NULL;
 	
-	gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &end, text, len);
-}
-
-static GeditDocumentNewlineType
-get_newline_type (GtkTextIter *end)
-{
-	GeditDocumentNewlineType res;
-	GtkTextIter copy;
-	gunichar c;
-
-	copy = *end;
-	c = gtk_text_iter_get_char (&copy);
-
-	GtkTextIter tt = copy;
-	gtk_text_iter_forward_chars (&tt, 2);
-
-	if (g_unichar_break_type (c) == G_UNICODE_BREAK_CARRIAGE_RETURN)
+	/* check cancelled state manually */
+	if (g_cancellable_is_cancelled (async->cancellable))
 	{
-		if (gtk_text_iter_forward_char (&copy) &&
-		    g_unichar_break_type (gtk_text_iter_get_char (&copy)) == G_UNICODE_BREAK_LINE_FEED)
-		{
-			res = GEDIT_DOCUMENT_NEWLINE_TYPE_CR_LF;
-		}
-		else
-		{
-			res = GEDIT_DOCUMENT_NEWLINE_TYPE_CR;
-		}
+		async_data_free (async);
+		return;
 	}
-	else
+	
+	gedit_debug_message (DEBUG_SAVER, "Finished closing input stream");
+	
+	if (!g_input_stream_close_finish (stream, res, &error))
 	{
-		res = GEDIT_DOCUMENT_NEWLINE_TYPE_LF;
+		gedit_debug_message (DEBUG_SAVER, "Closing input stream error: %s", error->message);
+
+		async_failed (async, error);
+		return;
 	}
 
-	return res;
+	gedit_debug_message (DEBUG_SAVER, "Close output stream");
+	if (!g_output_stream_close (async->loader->priv->output,
+				    async->cancellable, &error))
+	{
+		async_failed (async, error);
+		return;
+	}
+
+	remote_load_completed_or_failed (async->loader, async);
 }
 
 static void
-detect_newline_type (GeditDocumentLoader *loader)
+write_complete (AsyncData *async)
 {
-	GtkTextIter iter;
+	GeditDocumentLoader *loader;
 
-	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (loader->document), &iter);
+	loader = GEDIT_DOCUMENT_LOADER (async->loader);
 
-	if (!gtk_text_iter_backward_line (&iter))
-	{
-		gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (loader->document), &iter);
-	}
-
-	if (gtk_text_iter_ends_line (&iter) || gtk_text_iter_forward_to_line_end (&iter))
-	{
-		loader->auto_detected_newline_type = get_newline_type (&iter);
-	}
+	if (async->loader->priv->stream)
+		g_input_stream_close_async (G_INPUT_STREAM (async->loader->priv->stream),
+					    G_PRIORITY_HIGH,
+					    async->cancellable,
+					    (GAsyncReadyCallback)close_input_stream_ready_cb,
+					    async);
 }
 
+/* prototype, because they call each other... isn't C lovely */
+static void	read_file_chunk		(AsyncData *async);
+
 static void
-remove_ending_newline (GeditDocumentLoader *loader)
+write_file_chunk (AsyncData *async)
 {
-	GtkTextIter end;
-	GtkTextIter start;
+	GeditGioDocumentLoader *gvloader;
+	gssize bytes_written;
+	GError *error = NULL;
 
-	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (loader->document), &end);
-	start = end;
+	gvloader = async->loader;
 
-	gtk_text_iter_set_line_offset (&start, 0);
+	/* we use sync methods on doc stream since it is in memory. Using async
+	   would be racy and we can endup with invalidated iters */
+	bytes_written = g_output_stream_write (G_OUTPUT_STREAM (gvloader->priv->output),
+					       gvloader->priv->buffer,
+					       async->read,
+					       async->cancellable,
+					       &error);
 
-	if (gtk_text_iter_ends_line (&start) &&
-	    gtk_text_iter_backward_line (&start))
+	gedit_debug_message (DEBUG_SAVER, "Written: %" G_GSSIZE_FORMAT, bytes_written);
+	if (bytes_written == -1)
 	{
-		if (!gtk_text_iter_ends_line (&start))
-		{
-			gtk_text_iter_forward_to_line_end (&start);
-		}
-
-		/* Delete the empty line which is from 'start' to 'end' */
-		gtk_text_buffer_delete (GTK_TEXT_BUFFER (loader->document),
-		                        &start,
-		                        &end);
+		gedit_debug_message (DEBUG_SAVER, "Write error: %s", error->message);
+		async_failed (async, error);
+		return;
 	}
-}
 
-static void
-end_append_text_to_document (GeditDocumentLoader *loader)
-{
-	detect_newline_type (loader);
+	/* note that this signal blocks the read... check if it isn't
+	 * a performance problem
+	 */
+	gedit_document_loader_loading (GEDIT_DOCUMENT_LOADER (gvloader),
+				       FALSE,
+				       NULL);
 
-	/* If the last char is a newline, remove it from the buffer (otherwise
-		GtkTextView shows it as an empty line). See bug #324942. */
-	remove_ending_newline (loader);
-
-	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (loader->document), FALSE);
-
-	gtk_source_buffer_end_not_undoable_action (GTK_SOURCE_BUFFER (loader->document));
-
-	GEDIT_GIO_DOCUMENT_LOADER (loader)->priv->started_insert = FALSE;
+	read_file_chunk (async);
 }
 
 static void
@@ -371,7 +344,6 @@ async_read_cb (GInputStream *stream,
 	gedit_debug (DEBUG_LOADER);
 	GeditGioDocumentLoader *gvloader;
 	GeditDocumentLoader *loader;
-	gssize bytes_read;
 	GError *error = NULL;
 
 	gvloader = async->loader;
@@ -380,44 +352,43 @@ async_read_cb (GInputStream *stream,
 	/* manually check cancelled state */
 	if (g_cancellable_is_cancelled (async->cancellable))
 	{
-		end_append_text_to_document (GEDIT_DOCUMENT_LOADER (gvloader));
 		remote_load_completed_or_failed (gvloader, async);
 		return;
 	}
 
-	bytes_read = g_input_stream_read_finish (stream, res, &error);
+	async->read = g_input_stream_read_finish (stream, res, &error);
 	
 	/* error occurred */
-	if (bytes_read == -1)
+	if (async->read == -1)
 	{
 		async_failed (async, error);
 		return;
 	}
 
 	/* Check for the extremely unlikely case where the file size overflows. */
-	if (gvloader->priv->bytes_read + bytes_read < gvloader->priv->bytes_read)
+	if (gvloader->priv->bytes_read + async->read < gvloader->priv->bytes_read)
 	{
 		g_set_error (&gvloader->priv->error,
 			     GEDIT_DOCUMENT_ERROR,
 			     GEDIT_DOCUMENT_ERROR_TOO_BIG,
 			     "File too big");
 
-		end_append_text_to_document (GEDIT_DOCUMENT_LOADER (gvloader));
-		remote_load_completed_or_failed (gvloader, async);
-
+		async_failed (async, gvloader->priv->error);
 		return;
 	}
 
 	/* Bump the size. */
-	gvloader->priv->bytes_read += bytes_read;
+	gvloader->priv->bytes_read += async->read;
 
 	/* end of the file, we are done! */
-	if (bytes_read == 0)
+	if (async->read == 0)
 	{
 		GEDIT_DOCUMENT_LOADER (gvloader)->auto_detected_encoding =
 			gedit_smart_charset_converter_get_guessed (gvloader->priv->converter);
 
 		loader->auto_detected_encoding = gedit_smart_charset_converter_get_guessed (gvloader->priv->converter);
+		loader->auto_detected_newline_type =
+			gedit_document_output_stream_detect_newline_type (GEDIT_DOCUMENT_OUTPUT_STREAM (gvloader->priv->output));
 
 		/* Check if we needed some fallback char, if so, check if there was
 		   a previous error and if not set a fallback used error */
@@ -432,27 +403,12 @@ async_read_cb (GInputStream *stream,
 					     "needed to use a fallback char");
 		}*/
 
-		end_append_text_to_document (GEDIT_DOCUMENT_LOADER (gvloader));
-
-		remote_load_completed_or_failed (gvloader, async);
+		write_complete (async);
 
 		return;
 	}
 
-	append_text_to_document (GEDIT_DOCUMENT_LOADER (gvloader),
-				 gvloader->priv->buffer,
-				 bytes_read);
-
-	/* otherwise emit progress and read some more */
-
-	/* note that this signal blocks the read... check if it isn't
-	 * a performance problem
-	 */
-	gedit_document_loader_loading (GEDIT_DOCUMENT_LOADER (gvloader),
-				       FALSE,
-				       NULL);
-
-	read_file_chunk (async);
+	write_file_chunk (async);
 }
 
 static void
@@ -462,20 +418,6 @@ read_file_chunk (AsyncData *async)
 	
 	gvloader = async->loader;
 
-	if (!gvloader->priv->started_insert)
-	{
-		GeditDocumentLoader *loader;
-
-		loader = GEDIT_DOCUMENT_LOADER (gvloader);
-
-		/* Init the undoable action */
-		gtk_source_buffer_begin_not_undoable_action (GTK_SOURCE_BUFFER (loader->document));
-		gvloader->priv->started_insert = TRUE;
-
-		/* clear the buffer */
-		gtk_text_buffer_set_text (GTK_TEXT_BUFFER (loader->document), "", 0);
-	}
-
 	g_input_stream_read_async (G_INPUT_STREAM (gvloader->priv->stream),
 				   gvloader->priv->buffer,
 				   READ_CHUNK_SIZE,
@@ -552,6 +494,9 @@ finish_query_info (AsyncData *async)
 
 	gvloader->priv->stream = utf8_stream;
 
+	/* Output stream */
+	gvloader->priv->output = gedit_document_output_stream_new (loader->document);
+
 	/* start reading */
 	read_file_chunk (async);
 }
