@@ -37,19 +37,22 @@
 #include <gio/gio.h>
 #include <string.h>
 
-#include "gedit-convert.h"
 #include "gedit-gio-document-saver.h"
+#include "gedit-document-input-stream.h"
 #include "gedit-debug.h"
 
+#define WRITE_CHUNK_SIZE 8192
+
 typedef struct
 {
 	GeditGioDocumentSaver *saver;
-	gchar 		      *buffer;
+	gchar 		       buffer[WRITE_CHUNK_SIZE];
 	GCancellable 	      *cancellable;
 	gboolean	       tried_mount;
+	gssize		       written;
+	gssize		       read;
 } AsyncData;
 
-#define WRITE_CHUNK_SIZE 8192
 #define REMOTE_QUERY_ATTRIBUTES G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE "," \
 				G_FILE_ATTRIBUTE_TIME_MODIFIED
 				
@@ -58,9 +61,7 @@ typedef struct
 							  GeditGioDocumentSaverPrivate))
 
 static void	     gedit_gio_document_saver_save		    (GeditDocumentSaver *saver,
-								     time_t              old_mtime);
-static const gchar *gedit_gio_document_saver_get_content_type	    (GeditDocumentSaver *saver);
-static time_t       gedit_gio_document_saver_get_mtime		    (GeditDocumentSaver *saver);
+								     GTimeVal           *old_mtime);
 static goffset	     gedit_gio_document_saver_get_file_size	    (GeditDocumentSaver *saver);
 static goffset	     gedit_gio_document_saver_get_bytes_written	    (GeditDocumentSaver *saver);
 
@@ -69,15 +70,15 @@ static void 	    check_modified_async 			    (AsyncData          *async);
 
 struct _GeditGioDocumentSaverPrivate
 {
-	time_t                    doc_mtime;
-	gchar                    *content_type;
+	GTimeVal		  old_mtime;
 
 	goffset			  size;
 	goffset			  bytes_written;
 
 	GFile			 *gfile;
 	GCancellable		 *cancellable;
-	GFileOutputStream	 *stream;
+	GOutputStream		 *stream;
+	GInputStream		 *input;
 	
 	GError                   *error;
 };
@@ -85,28 +86,36 @@ struct _GeditGioDocumentSaverPrivate
 G_DEFINE_TYPE(GeditGioDocumentSaver, gedit_gio_document_saver, GEDIT_TYPE_DOCUMENT_SAVER)
 
 static void
-gedit_gio_document_saver_finalize (GObject *object)
+gedit_gio_document_saver_dispose (GObject *object)
 {
 	GeditGioDocumentSaverPrivate *priv = GEDIT_GIO_DOCUMENT_SAVER (object)->priv;
 
-	if (priv->cancellable)
+	if (priv->cancellable != NULL)
 	{
 		g_cancellable_cancel (priv->cancellable);
 		g_object_unref (priv->cancellable);
+		priv->cancellable = NULL;
 	}
 
-	if (priv->gfile)
+	if (priv->gfile != NULL)
+	{
 		g_object_unref (priv->gfile);
+		priv->gfile = NULL;
+	}
 
-	g_free (priv->content_type);
-
-	if (priv->error)
+	if (priv->error != NULL)
+	{
 		g_error_free (priv->error);
-		
-	if (priv->stream)
-		g_object_unref (priv->stream);
+		priv->error = NULL;
+	}
 
-	G_OBJECT_CLASS (gedit_gio_document_saver_parent_class)->finalize (object);
+	if (priv->stream != NULL)
+	{
+		g_object_unref (priv->stream);
+		priv->stream = NULL;
+	}
+
+	G_OBJECT_CLASS (gedit_gio_document_saver_parent_class)->dispose (object);
 }
 
 static AsyncData *
@@ -116,9 +125,10 @@ async_data_new (GeditGioDocumentSaver *gvsaver)
 	
 	async = g_slice_new (AsyncData);
 	async->saver = gvsaver;
-	async->buffer = NULL;
 	async->cancellable = g_object_ref (gvsaver->priv->cancellable);
 	async->tried_mount = FALSE;
+	async->written = 0;
+	async->read = 0;
 	
 	return async;
 }
@@ -127,7 +137,6 @@ static void
 async_data_free (AsyncData *async)
 {
 	g_object_unref (async->cancellable);
-	g_free (async->buffer);
 	g_slice_free (AsyncData, async);
 }
 
@@ -137,11 +146,9 @@ gedit_gio_document_saver_class_init (GeditGioDocumentSaverClass *klass)
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 	GeditDocumentSaverClass *saver_class = GEDIT_DOCUMENT_SAVER_CLASS (klass);
 
-	object_class->finalize = gedit_gio_document_saver_finalize;
+	object_class->finalize = gedit_gio_document_saver_dispose;
 
 	saver_class->save = gedit_gio_document_saver_save;
-	saver_class->get_content_type = gedit_gio_document_saver_get_content_type;
-	saver_class->get_mtime = gedit_gio_document_saver_get_mtime;
 	saver_class->get_file_size = gedit_gio_document_saver_get_file_size;
 	saver_class->get_bytes_written = gedit_gio_document_saver_get_bytes_written;
 
@@ -178,38 +185,10 @@ async_failed (AsyncData *async,
 	remote_save_completed_or_failed (async->saver, async);
 }
 
-/* prototype, because they call each other... isn't C lovely */
-static void write_file_chunk (AsyncData *async);
-
 static void
-handle_file_info (AsyncData *async,
-		  GFileInfo *info)
-{
-	GeditGioDocumentSaver *saver;
-	GTimeVal mtime;
-
-	saver = async->saver;
-
-	gedit_debug_message (DEBUG_SAVER, "Query info finished");
-
-	if (g_file_info_has_attribute (info,
-				       G_FILE_ATTRIBUTE_TIME_MODIFIED))
-	{
-		g_file_info_get_modification_time (info, &mtime);
-		saver->priv->doc_mtime = mtime.tv_sec;
-	}
-
-	if (g_file_info_has_attribute (info,
-				       G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE))
-	{
-		saver->priv->content_type = g_strdup (g_file_info_get_content_type (info));
-	}
-}
-
-static void
-remote_reget_info_cb (GFile        *source,
-		      GAsyncResult *res,
-		      AsyncData    *async)
+remote_get_info_cb (GFile        *source,
+		    GAsyncResult *res,
+		    AsyncData    *async)
 {
 	GeditGioDocumentSaver *saver;
 	GFileInfo *info;
@@ -229,8 +208,10 @@ remote_reget_info_cb (GFile        *source,
 
 	if (info != NULL)
 	{
-		handle_file_info (async, info);
-		g_object_unref (info);
+		if (GEDIT_DOCUMENT_SAVER (saver)->info != NULL)
+			g_object_unref (GEDIT_DOCUMENT_SAVER (saver)->info);
+
+		GEDIT_DOCUMENT_SAVER (saver)->info = info;
 	}
 	else
 	{
@@ -264,135 +245,67 @@ close_async_ready_get_info_cb (GOutputStream *stream,
 		async_failed (async, error);
 		return;
 	}
-	
+
+	/* get the file info: note we cannot use 
+	 * g_file_output_stream_query_info_async since it is not able to get the
+	 * content type etc, beside it is not supported by gvfs.
+	 * I'm not sure this is actually necessary, can't we just use
+	 * g_content_type_guess (since we have the file name and the data)
+	 */
 	gedit_debug_message (DEBUG_SAVER, "Query info on file");
 	g_file_query_info_async (async->saver->priv->gfile,
 			         REMOTE_QUERY_ATTRIBUTES,
 			         G_FILE_QUERY_INFO_NONE,
 			         G_PRIORITY_HIGH,
 			         async->cancellable,
-			         (GAsyncReadyCallback) remote_reget_info_cb,
+			         (GAsyncReadyCallback) remote_get_info_cb,
 			         async);
 }
 
-static void
-close_async_ready_cb (GOutputStream *stream,
-		      GAsyncResult  *res,
-		      AsyncData     *async)
-{
-	GError *error = NULL;
-
-	/* check cancelled state manually */
-	if (g_cancellable_is_cancelled (async->cancellable))
-	{
-		async_data_free (async);
-		return;
-	}
-
-	if (!g_output_stream_close_finish (stream, res, &error))
-	{
-		g_propagate_error (&async->saver->priv->error, error);
-	}
-
-	remote_save_completed_or_failed (async->saver, async);
-}
-
-static void
-remote_get_info_cb (GFileOutputStream *stream,
-		    GAsyncResult      *res,
-		    AsyncData         *async)
-{
-	GeditGioDocumentSaver *saver;
-	GFileInfo *info;
-	GError *error = NULL;
-	GAsyncReadyCallback next_callback;
-
-	saver = async->saver;
-
-	/* check cancelled state manually */
-	if (g_cancellable_is_cancelled (async->cancellable))
-	{
-		async_data_free (async);
-		return;
-	}
-	
-	gedit_debug_message (DEBUG_SAVER, "Finishing info query");
-	info = g_file_output_stream_query_info_finish (stream, res, &error);
-	gedit_debug_message (DEBUG_SAVER, "Query info result: %s", info ? "ok" : "fail");
-
-	if (info == NULL)
-	{
-		if (error->code == G_IO_ERROR_NOT_SUPPORTED || error->code == G_IO_ERROR_CLOSED)
-		{
-			gedit_debug_message (DEBUG_SAVER, "Query info not supported on stream, trying on file");
-
-			/* apparently the output stream does not support query info.
-			 * we're forced to restat on the file. But first we make sure
-			 * to close the stream
-			 */
-			g_error_free (error);
-			next_callback = (GAsyncReadyCallback) close_async_ready_get_info_cb;
-		}
-		else
-		{
-			gedit_debug_message (DEBUG_SAVER, "Query info failed: %s", error->message);
-			g_propagate_error (&saver->priv->error, error);
-
-			next_callback = (GAsyncReadyCallback) close_async_ready_cb;
-		}
-	}
-	else
-	{
-		handle_file_info (async, info);
-		g_object_unref (info);
-
-		next_callback = (GAsyncReadyCallback) close_async_ready_cb;
-	}
-
-	g_output_stream_close_async (G_OUTPUT_STREAM (saver->priv->stream),
-				     G_PRIORITY_HIGH,
-				     async->cancellable,
-				     next_callback,
-				     async);
-}
-
 static void
 write_complete (AsyncData *async)
 {
-	/* document is succesfully saved. we know requery for the mime type and
-	 * the mtime. I'm not sure this is actually necessary, can't we just use
-	 * g_content_type_guess (since we have the file name and the data)
-	 */
-	gedit_debug_message (DEBUG_SAVER, "Write complete, query info on stream");
-	
-	/* query info on the stream */
-	g_file_output_stream_query_info_async (async->saver->priv->stream, 
-					       REMOTE_QUERY_ATTRIBUTES,
-					       G_PRIORITY_HIGH,
-					       async->cancellable,
-					       (GAsyncReadyCallback) remote_get_info_cb,
-					       async);
+	GError *error = NULL;
+
+	/* first we close the input stream */
+	gedit_debug_message (DEBUG_SAVER, "Close input stream");
+	if (!g_input_stream_close (async->saver->priv->input,
+				   async->cancellable, &error))
+	{
+		gedit_debug_message (DEBUG_SAVER, "Closing input stream error: %s", error->message);
+		async_failed (async, error);
+		return;
+	}
+
+	/* now we close the output stream */
+	gedit_debug_message (DEBUG_SAVER, "Close output stream");
+	g_output_stream_close_async (async->saver->priv->stream,
+				     G_PRIORITY_HIGH,
+				     async->cancellable,
+				     (GAsyncReadyCallback)close_async_ready_get_info_cb,
+				     async);
 }
 
+/* prototype, because they call each other... isn't C lovely */
+static void read_file_chunk (AsyncData *async);
+static void write_file_chunk (AsyncData *async);
+
 static void
 async_write_cb (GOutputStream *stream,
-	        GAsyncResult  *res,
-	        AsyncData     *async)
+		GAsyncResult  *res,
+		AsyncData     *async)
 {
-	GError *error = NULL;
 	GeditGioDocumentSaver *gvsaver;
 	gssize bytes_written;
-	
-	gedit_debug (DEBUG_SAVER);
-	
-	/* manually check cancelled state */
+	GError *error = NULL;
+
+	/* Check cancelled state manually */
 	if (g_cancellable_is_cancelled (async->cancellable))
 	{
 		async_data_free (async);
 		return;
 	}
 
-	gvsaver = async->saver;
 	bytes_written = g_output_stream_write_finish (stream, res, &error);
 
 	gedit_debug_message (DEBUG_SAVER, "Written: %" G_GSSIZE_FORMAT, bytes_written);
@@ -403,17 +316,16 @@ async_write_cb (GOutputStream *stream,
 		async_failed (async, error);
 		return;
 	}
-	
-	gvsaver->priv->bytes_written += bytes_written;
-	
-	/* if nothing is written we're done */
-	if (gvsaver->priv->bytes_written == gvsaver->priv->size)
+
+	gvsaver = async->saver;
+	async->written += bytes_written;
+
+	/* write again */
+	if (async->written != async->read)
 	{
-		write_complete (async);
+		write_file_chunk (async);
 		return;
 	}
-	
-	/* otherwise emit progress and write some more */
 
 	/* note that this signal blocks the write... check if it isn't
 	 * a performance problem
@@ -422,7 +334,7 @@ async_write_cb (GOutputStream *stream,
 				     FALSE,
 				     NULL);
 
-	write_file_chunk (async);
+	read_file_chunk (async);
 }
 
 static void
@@ -432,26 +344,62 @@ write_file_chunk (AsyncData *async)
 
 	gvsaver = async->saver;
 
-	gedit_debug_message (DEBUG_SAVER,
-			     "Writing next chunk: %" G_GINT64_FORMAT "/%" G_GINT64_FORMAT,
-			     gvsaver->priv->bytes_written, gvsaver->priv->size);
-
 	g_output_stream_write_async (G_OUTPUT_STREAM (gvsaver->priv->stream),
-				     async->buffer + gvsaver->priv->bytes_written,
-				     MIN (WRITE_CHUNK_SIZE, gvsaver->priv->size - 
-				     			    gvsaver->priv->bytes_written),
+				     async->buffer + async->written,
+				     async->read - async->written,
 				     G_PRIORITY_HIGH,
 				     async->cancellable,
 				     (GAsyncReadyCallback) async_write_cb,
 				     async);
 }
 
+static void
+read_file_chunk (AsyncData *async)
+{
+	GeditGioDocumentSaver *gvsaver;
+	GeditDocumentInputStream *dstream;
+	GError *error = NULL;
+
+	gvsaver = async->saver;
+	async->written = 0;
+
+	/* we use sync methods on doc stream since it is in memory. Using async
+	   would be racy and we can endup with invalidated iters */
+	async->read = g_input_stream_read (gvsaver->priv->input,
+					   async->buffer,
+					   WRITE_CHUNK_SIZE,
+					   async->cancellable,
+					   &error);
+
+	if (error != NULL)
+	{
+		async_failed (async, error);
+		return;
+	}
+
+	/* Check if we finished reading and writing */
+	if (async->read == 0)
+	{
+		write_complete (async);
+		return;
+	}
+
+	/* Get how many chars have been read */
+	dstream = GEDIT_DOCUMENT_INPUT_STREAM (gvsaver->priv->input);
+	gvsaver->priv->bytes_written = gedit_document_input_stream_tell (dstream);
+
+	write_file_chunk (async);
+}
+
 static void
 async_replace_ready_callback (GFile        *source,
 			      GAsyncResult *res,
 			      AsyncData    *async)
 {
 	GeditGioDocumentSaver *gvsaver;
+	GeditDocumentSaver *saver;
+	GCharsetConverter *converter;
+	GFileOutputStream *file_stream;
 	GError *error = NULL;
 
 	/* Check cancelled state manually */
@@ -462,71 +410,51 @@ async_replace_ready_callback (GFile        *source,
 	}
 	
 	gvsaver = async->saver;
-	gvsaver->priv->stream = g_file_replace_finish (source, res, &error);
+	saver = GEDIT_DOCUMENT_SAVER (gvsaver);
+	file_stream = g_file_replace_finish (source, res, &error);
 	
 	/* handle any error that might occur */
-	if (!gvsaver->priv->stream)
+	if (!file_stream)
 	{
 		gedit_debug_message (DEBUG_SAVER, "Opening file failed: %s", error->message);
 		async_failed (async, error);
 		return;
 	}
 
-	/* now that we have the stream, start writing async to it */
-	write_file_chunk (async);
-}
+	/* FIXME: manage converter error? */
+	gedit_debug_message (DEBUG_SAVER, "Encoding charset: %s",
+			     gedit_encoding_get_charset (saver->encoding));
 
-/* returns a pointer to the reallocated buffer */
-static gchar *
-append_new_line (GeditGioDocumentSaver  *gvsaver, 
-		 gchar 		        *buffer, 
-		 gsize 		        *len, 
-		 GError 	       **error)
-{
-	GeditDocumentSaver *saver = GEDIT_DOCUMENT_SAVER (gvsaver);
-	gsize n_len;
-	gchar *n_buffer;
-	gchar *res;
-
-	res = buffer;
-
-	n_buffer = gedit_document_saver_get_end_newline (saver, &n_len);
-	if (n_buffer != NULL)
+	if (saver->encoding != gedit_encoding_get_utf8 ())
 	{
-		gchar *new_buffer;
+		converter = g_charset_converter_new (gedit_encoding_get_charset (saver->encoding),
+						     "UTF-8",
+						     NULL);
+		gvsaver->priv->stream = g_converter_output_stream_new (G_OUTPUT_STREAM (file_stream),
+								       G_CONVERTER (converter));
 
-		new_buffer = g_try_realloc (buffer, *len + n_len + 1);
-
-		if (new_buffer == NULL)
-		{
-			/* we do not error out, just use the buffer
-			   without the ending newline */
-			g_free (n_buffer);
-			g_warning ("Cannot add '\\n' at the end of the file.");
-
-			return res;
-		}
-
-		/* Copy newline */
-		memcpy (new_buffer + *len, n_buffer, n_len);
-		g_free (n_buffer);
-		*len += n_len;
-
-		new_buffer[*len] = '\0';
-
-		res = new_buffer;
+		g_object_unref (file_stream);
+		g_object_unref (converter);
 	}
+	else
+	{
+		gvsaver->priv->stream = G_OUTPUT_STREAM (file_stream);
+	}
+	
+	gvsaver->priv->input = gedit_document_input_stream_new (GTK_TEXT_BUFFER (saver->document),
+								saver->newline_type);
 
-	return res;
+	gvsaver->priv->size = gedit_document_input_stream_get_total_size (GEDIT_DOCUMENT_INPUT_STREAM (gvsaver->priv->input));
+
+	read_file_chunk (async);
 }
 
 static void
 begin_write (AsyncData *async)
 {
 	GeditGioDocumentSaver *gvsaver;
-	gchar *buffer;
-	gsize len;
-	GError *error = NULL;
+	GeditDocumentSaver *saver;
+	gboolean backup;
 
 	gedit_debug_message (DEBUG_SAVER, "Start replacing file contents");
 
@@ -534,28 +462,18 @@ begin_write (AsyncData *async)
 	 * backup as of yet
 	 */
 	gvsaver = async->saver;
-	buffer = gedit_document_saver_get_document_contents (GEDIT_DOCUMENT_SAVER (gvsaver), &len, &error);
-	if (buffer != NULL && len > 0)
-	{
-		/* Append new line to buffer */
-		buffer = append_new_line (gvsaver, buffer, &len, &error);
-	}
+	saver = GEDIT_DOCUMENT_SAVER (gvsaver);
 
-	if (!buffer)
-	{
-		async_failed (async, error);
-		return;
-	}
-
-	async->buffer = buffer;
-	gvsaver->priv->size = len;
+	/* Do not make backups for remote files so they do not clutter remote systems */
+	backup = (saver->keep_backup && gedit_document_is_local (saver->document));
 
 	gedit_debug_message (DEBUG_SAVER, "File contents size: %" G_GINT64_FORMAT, gvsaver->priv->size);
 	gedit_debug_message (DEBUG_SAVER, "Calling replace_async");
+	gedit_debug_message (DEBUG_SAVER, backup ? "Keep backup" : "Discard backup");
 
-	g_file_replace_async (gvsaver->priv->gfile, 
+	g_file_replace_async (gvsaver->priv->gfile,
 			      NULL,
-			      FALSE,
+			      backup,
 			      G_FILE_CREATE_NONE,
 			      G_PRIORITY_HIGH,
 			      async->cancellable,
@@ -608,9 +526,9 @@ recover_not_mounted (AsyncData *async)
 				       mount_operation,
 				       async->cancellable,
 				       (GAsyncReadyCallback) mount_ready_callback,
-				       async);	
+				       async);
 
-	g_object_unref (mount_operation);			
+	g_object_unref (mount_operation);
 }
 
 static void
@@ -655,11 +573,13 @@ check_modification_callback (GFile        *source,
 				       G_FILE_ATTRIBUTE_TIME_MODIFIED))
 	{
 		GTimeVal mtime;
+		GTimeVal old_mtime;
 
 		g_file_info_get_modification_time (info, &mtime);
+		old_mtime = gvsaver->priv->old_mtime;
 
-		if (gvsaver->priv->doc_mtime > 0 &&
-		    mtime.tv_sec != gvsaver->priv->doc_mtime &&
+		if ((old_mtime.tv_sec > 0 || old_mtime.tv_usec > 0) &&
+		    (mtime.tv_sec != old_mtime.tv_sec || mtime.tv_usec != old_mtime.tv_usec) &&
 		    (GEDIT_DOCUMENT_SAVER (gvsaver)->flags & GEDIT_DOCUMENT_SAVE_IGNORE_MTIME) == 0)
 		{
 			gedit_debug_message (DEBUG_SAVER, "File is externally modified");
@@ -716,11 +636,11 @@ save_remote_file_real (GeditGioDocumentSaver *gvsaver)
 
 static void
 gedit_gio_document_saver_save (GeditDocumentSaver *saver,
-			       time_t              old_mtime)
+			       GTimeVal           *old_mtime)
 {
 	GeditGioDocumentSaver *gvsaver = GEDIT_GIO_DOCUMENT_SAVER (saver);
 
-	gvsaver->priv->doc_mtime = old_mtime;
+	gvsaver->priv->old_mtime = *old_mtime;
 	gvsaver->priv->gfile = g_file_new_for_uri (saver->uri);
 
 	/* saving start */
@@ -733,18 +653,6 @@ gedit_gio_document_saver_save (GeditDocumentSaver *saver,
 			    NULL);
 }
 
-static const gchar *
-gedit_gio_document_saver_get_content_type (GeditDocumentSaver *saver)
-{
-	return GEDIT_GIO_DOCUMENT_SAVER (saver)->priv->content_type;
-}
-
-static time_t
-gedit_gio_document_saver_get_mtime (GeditDocumentSaver *saver)
-{
-	return GEDIT_GIO_DOCUMENT_SAVER (saver)->priv->doc_mtime;
-}
-
 static goffset
 gedit_gio_document_saver_get_file_size (GeditDocumentSaver *saver)
 {
