@@ -4,7 +4,7 @@ import sys
 import shutil
 
 from snippets.Library import *
-import ElementTree as et
+import xml.etree.ElementTree as et
 from Helper import *
 
 class Exporter:
@@ -62,7 +62,7 @@ class Exporter:
                         os.chdir(curdir)
                 
                 if status != 0:
-                        return _('The archive `%s` could not be created' % self.filename)
+                        return _('The archive "%s" could not be created' % self.filename)
                 
                 # Remove the temporary directory
                 shutil.rmtree(dirname)
@@ -79,10 +79,10 @@ class Exporter:
         def run(self):
                 dirname = os.path.dirname(self.filename)
                 if not os.path.exists(dirname):
-                        return _('Target directory `%s` does not exist') % dirname
+                        return _('Target directory "%s" does not exist') % dirname
                 
                 if not os.path.isdir(dirname):
-                        return _('Target directory `%s` is not a valid directory') % dirname
+                        return _('Target directory "%s" is not a valid directory') % dirname
                 
                 (root, ext) = os.path.splitext(self.filename)
                 
